#ifndef __DOCRA_PARSER_HPP
#define __DOCRA_PARSER_HPP

#include <list>
#include <map>
#include <regex>
#include <sstream>
#include <unordered_map>

#include "lexer.hpp"
#include "util.hpp"
#include "fs.hpp"

namespace docra {//docra``+

  //+`ConfigEntry`
  typedef std::pair<std::string, std::string> ConfigEntry;

  //+`ConfigEntries`
  typedef std::list<ConfigEntry> ConfigEntries;

  //+`ConfigSubSections`
  typedef std::unordered_map<std::string, ConfigEntries> ConfigSubSections;

  //+`ConfigSections`
  typedef std::unordered_map<std::string, ConfigSubSections> ConfigSections;

  //+`Config`
  struct Config
  {
    fs::path base_dir;

    int cli_verbose;

//ug`config.syntax`common.verbose _ verbose = _verblevel_:: The global verbosity level, identified by an integer. The following levels are defined, in increasing order of verbosity:
// ** 0 : Display critical errors only
// ** 1 : Display warnings
// ** 2 : Display information messages (default)
// ** 3 : Display extra information for debugging
// ** 4 : Trace internal state of each lexer and parser operation

    enum {
      VERB_CRIT = 0, VERB_WARN, VERB_INFO, VERB_DEBUG, VERB_INTERNAL
    };
    std::string cli_outdir;
    std::string cli_urlprefix;
    std::string cli_urlsuffix;
    int cli_level;
    bool cli_pathsonly;

    ConfigSections sections;

    Filters common_filters;
    Filters lex_clike_filters;
    Filters lex_hash_filters;
    Filters lex_vhdl_filters;
    Filters lex_ini_filters;
    Filters lex_adoc_filters;
    Filters lex_yacc_filters;
    Filters lex_xml_filters;

    static void exclude(Filters &filters, const std::string &spec);
    static void include(Filters &filters, const std::string &spec);

    Config();

    void reset();

    void apply(const std::string &sec, const std::string &sub, Filters &filters);
    void apply();

    static ConfigEntries::reverse_iterator find(ConfigEntries &entries, const std::string &str);

    int verbose(const std::string &sec = "common", const std::string &sub = "");

    fs::path fullpath(const fs::path &path)
    {
      return path.is_absolute() ? path : base_dir / path;
    }

    int path_level(const fs::path &sub);
    std::string path_url(const fs::path &sub);

    fs::path template_outdir(const fs::path &sub);
    fs::path template_outfile(const fs::path &sub);

    //std::string curr_sec;
    //std::string curr_sub;

    //void select(const std::string &sec, const std::string &sub)
    //{
    //  curr_sec = sec; curr_sub = sub;
    //}

    bool has_section(const std::string &sec) const
    {
      return sections.find(sec) != sections.end();
    }

    const ConfigSubSections &section(const std::string &sec)
    {
      return sections[sec];
    }

    void set(const std::string &sec, const std::string &sub);

    //++`::set`
    void set(const std::string &sec, const std::string &sub, const std::string &key, const std::string &val, bool replace);//?>

    template <typename K>
    void set(const std::string &sec, const std::string &sub, const std::string &key, K val, bool replace)
    {
      std::stringstream ss;
      ss << val;
      set(sec, sub, key, ss.str(), replace);
    }

    //++`::get` <?
    const ConfigEntries &get(const std::string &sec, const std::string &sub);//?>

    //++`::begin`
    //-> iterator over all sections called sec
    ConfigSections::iterator begin(const std::string &sec)//;?>
    {
      return sections.find(sec);
    }

    //++`::end`
    ConfigSections::iterator end()//;?>
    {
      return sections.end();
    }
  };

  //+`Location`
  struct Location {
    fs::path path;
    int lineno;

    Location() : lineno(0) {}
    Location(const fs::path &p, int l) : path(p), lineno(l) {}
  };

  //+`operator == (const Location&)` Comparison operator for source locations
  //>`a` left-hand operand
  //>`b` right-hand operand
  //-> boolean result of comparison if both operands point to same source location
  inline bool operator == (const Location &a, const Location &b)//;?>
  {
    return a.path == b.path && a.lineno == b.lineno;
  }

  struct Scope;
  typedef std::list<Scope> Scopes;

  struct Suffix;

  //+`Scope` Representation of a generic lexical scope within a `Project`
  struct Scope {
    Suffix *suffix; //++`.suffix` ?>Points to the suffix corresponding to the scope tag (optional)
    int lineno_start; //++`.lineno_start` ?>Source location of scope start
    int lineno_end;   //++`.lineno_end` ?>Source location of scope end

    Scope* parent;   //++`.parent` ?>Pointer to `Scope` containing this one
    Scopes children; //++`.children` ?>List of scopes semantically contained in this `Scope`

    Scope() : suffix(nullptr), lineno_start(0), lineno_end(0), parent(nullptr) {}

    Scope* add_child(int lineno); //++`::add_child` ?>Insert a new scope below the current hierarchy

    void print(size_t level = 0); //++`::print`
  };

  //+`FileScopes` Map from a file path to the root of its scope tree (populated by `Parser::parse` )
  typedef std::map<fs::path, Scope> FileScopes;//?>

  struct Project;

  //+`Parser_Source` Takes a stream of `Token` from `Lexer` to create a parse tree
  struct Parser_Source {
    Lexer &lexer;     //++`.lexer` ?>Reference to the current `Lexer`
    Project &project; //++`.project` ?>Reference to the current `Project`

    Parser_Source(Lexer &l, Project &p);

    bool parse(int verb, const fs::path &, Scope &); //++`::parse` ?>
  };

  struct Param { //+`Param`
    enum KIND { //++`::KIND`
      IN,       //++`.IN` Starts with ">"
      OUT,      //++`.OUT` Starts with "<"
      INOUT,    //++`.INOUT` Starts with "<>"
      GENERIC,  //++`.GENERIC` Starts with "="
      RET,      //++`.RET` Starts with "->"
      LOCAL     //++`.LOCAL` Starts with "-"
    };
    KIND kind;
    std::string name;  //++`.name` ?>Name of this parameter
    int lineno;
    Tokens desc; //++`.desc` ?>List of tokens describing this parameter
    Tokens desc_code; //++`.desc_code` ?>Code block describing this parameter
    bool inlined; //++`.inlined` ?>This param is listed in a group (and should not be rendered by itself but inside the group)

    Param(KIND k, std::string n, int l) : kind(k), name(n), lineno(l), inlined(false) {}

    void render(const Config &cfg, const Suffix &suffix, std::ostream &out) const;
  };
  typedef std::list<Param> Params;

  //+`Suffix` All data attached to a given (prefix, id, suffix) tuple
  struct Suffix {
    Project *project;//++`.project` ?>Project containing this tag
    Tag tag;         //++`.tag` ?>The full resolved tag identifying this suffix
    Token token_tag; //++`.token_tag` ?>A copy of the original suffix tag
    fs::path path;   //++`.path` ?>Path to file that this tag belongs to
    int lineno_code; //++`.lineno_code` ?>If this tag refers to code, the corresponding line number
    Scope *scope;    //++`.scope` ?>Pointer into the scope tree to where this suffix was defined
    Params params;   //++`.params` ?>Local symbols valid within the scope of this full tag
    int leveloffset;

    Tokens desc_code;  //++`.desc_code` ?>Code block describing this entity
    Tokens desc_short; //++`.desc_short` ?>Short description of this entity
    Tokens desc_long;  //++`.desc_long` ?>Long description of this entity

    //++`.lineno_from`
    int lineno_from() const//;?>
    {
      return token_tag.lineno;
    }
    int lineno_to();

    void render_pilcrow(const Config &cfg, std::ostream &out, int lineno) const;

    //++`.render_description`
    bool render_description(const Config &cfg, std::ostream &out, const Tokens &desc) const;//?>

    //++`.render_codeblock`
    bool render_codeblock(std::ostream &out, const Tokens &desc, Tokens::const_iterator &it) const;//?>
    //>`out` Stream to write to
    //><`it` Iterator to start of block

    bool render(const Config &cfg, std::ostream &out) const;

    //++`.resolve`
    Suffix *resolve(const Tag &in, Tag &out) const;

    const Param *find_param(const std::string &) const;

    Suffix()
    : lineno_code(0), scope(nullptr), leveloffset(0) {}

    Suffix(Project *p, const Tag &t, Token &&tok)
    : project(p), tag(t), token_tag(std::move(tok)), lineno_code(0), scope(nullptr), leveloffset(0) {}
  };
  typedef std::multimap<std::string, Suffix> Suffixes;

  //+`Symbol` All data assigned to a given tag id
  struct Symbol {
    Suffixes suffixes; //++`.suffixes` Map from an optional suffix to the associated `Suffix` data

    std::list<Suffixes::const_iterator> suffixes_unsorted; //++`.unsorted_suffixes`
    //? All the suffixes, but appearing in the same order they were inserted

    //Symbol *parent;               //++`.parent` Parent of this identifier, in the prefix hierarchy
    //std::list<Symbol *> children; //++`.children` All identifiers immediately below this one, in the hierarchy

    //++`.render_xref`
    bool render_xref(const Config &cfg, std::ostream &out) const;//?>

    //++`.render_xref_yacc`
    bool render_xref_yacc(const Config &cfg, std::ostream &out) const;//?>

    // Note: all suffixes attached to a given id should belong to the same hierarchy (the converse can happen in degenerate cases that we do not consider to be valid inputs)
  };
  typedef std::map<std::string, Symbol> Symbols;

  //+`Prefix` Document-wide state for a given prefix
  struct Prefix {
    Symbols symbols; //++`.symbols` Map from prefixed symbol name to all its existing suffixes

    //std::stack<Symbol *> hierarchy; //++`.hierarchy` Current stack of identifiers ordered by depth, used for grouping
    //std::list<Symbol *> top; //++`.top` Identifiers at the top of the hierarchy for this prefix
  };
  typedef std::map<std::string, Prefix> Prefixes;

  //+`Parser_Template`
  struct Parser_Template
  {
    Lexer_Adoc &lexer;
    Project &project;
    std::ostream &out;

    Parser_Template(Lexer_Adoc &l, Project &p, std::ostream &o);

    bool parse(int verb);

  private:
    bool next_token();
    const Token &current_token();
    std::smatch match;
    std::string::const_iterator it;
    Token nested_token;
  };

  //+`Parser_Config`
  struct Parser_Config
  {
    Config &_cfg;

    Parser_Config(Config &cfg) : _cfg(cfg) {}

    bool parse(const fs::path &path);
  };

} //namespace docra

#endif//__DOCRA_PARSER_HPP
