#include "util.hpp"

#include <algorithm>
#include "fs.hpp"

//ug`config.patterns` _ ==== Patterns
// Patterns can be used in a configuration file to select specific sets of files or directories. Two pattern syntaxes are available:
// _ Globbing syntax::
// In this case only two special globbing characters are supported, ``*`` matching any number of characters and ``?`` matching any single character.
// [example]
// ====
// For example, a pattern like ``*.adoc`` can be used to match any AsciiDoc file. Something like ``version?.log`` will match ``version1.log``, ``versionb.log`` and so on.
// ====
// _ Regular expression syntax::
// Given the limitations of the above syntax, it is possible to use full regular expressions (as implemented in the http://en.cppreference.com/w/cpp/regex[C++ standard library]) by starting the pattern with a ``$`` sign (the dollar will not be part of the regular expression).
// [example]
// ====
// For example, a pattern like ``$^.*/?[.]git/$`` can be used to match a ``.git`` folder anywhere in the hierarchy being traversed. Note that the terminating dollar _is_ part of the regular expression (as it matches the end of the string). Also note that this could not be accomplished just with globbing, as that would have matched any file or directory whose name ended in ``.git``.
// ====
// The second syntax is clearly more powerful but globbing is simpler and easier to read and should be preferred when possible.
// In these patterns, ``/`` should be used as the path separator. When running on Windows, those forward slashes will be automatically converted to backslashes, so that patterns will behave in the same way across platforms.

//ug`template.patterns` _ ==== Patterns
// Patterns can be used to control which tags are selected by template directives like ``?includexref``.
// [example]
// ====
// Considert the following two rules: _
// . ``${caret}(TODO|FIXME|BUG)${backtick}{backtick}``
// . ``${caret}(TODO|FIXME|BUG)${backtick}${caret}.+${backtick}``
// _ The first rule only includes tags under the TODO, FIXME and BUG prefix with an empty symbol name. The second only those with a nonempty symbol name.
// ====
// This syntax is identical to what is used in configuration file patterns, with the only difference that in this case path separators are not touched. Both the tag prefix and the symbol are regular expressions, the symbol part is surrounded by backticks following the usual tag syntax, a regular expression used inside a tag selector should not contain backticks.

//`make_regex`
std::regex docra::make_regex(const std::string &pattern, bool ispath)
{
#ifdef _WIN32
  std::string patt;
  if (ispath) {
    patt = std::regex_replace(pattern, std::regex("/"), "\\\\");
  } else {
    patt = pattern;
  }
#else
  std::string patt(pattern);
#endif
  if (patt[0] == '$') {
    return std::regex(patt.substr(1));
  } else {
    if (ispath && patt.length() && *patt.rbegin() == fs::path::preferred_separator) {
      patt += "*";
    }
    static const std::regex re1("\\.");
    static const std::regex re2("\\*");
    static const std::regex re3("\\?");
    std::string res = std::regex_replace(patt, re1, "[.]");
    res = std::regex_replace(res, re2, ".*");
    res = std::regex_replace(res, re3, ".");
    return std::regex("^" + res + "$");
  }
}
